from imutils import paths
import numpy as np
import os
import cv2
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, MaxPool2D,Flatten,Dense
from tensorflow.keras.optimizers import SGD
from keras.utils import to_categorical
import random
from tensorflow.keras.datasets import mnist

def load_test_data():
    """โหลดข้อมูล MNIST test dataset"""
    (_, _), (x_test, y_test) = mnist.load_data()
    # ปรับขนาดข้อมูลให้เหมาะสม
    x_test = x_test.astype('float32') / 255.0
    x_test = np.expand_dims(x_test, axis=-1)  # เพิ่มมิติให้ข้อมูล
    y_test = np.eye(10)[y_test]  # One-hot encoding สำหรับ labels
    return x_test, y_test

def load_mnist_byCid(cid):
    data = []
    labels = []

    path = f"data/client_{cid}"
    img_paths = list(paths.list_images(path))

    for imgpath in img_paths:
        img_grayscale = cv2.imread(imgpath, cv2.IMREAD_GRAYSCALE)
        img_resized = cv2.resize(img_grayscale, (28, 28))
        img = np.expand_dims(img_resized, axis=-1) / 255.0
        label = imgpath.split(os.path.sep)[-2]
        data.append(img)
        labels.append(label)
    return np.array(data), np.array(labels)

def define_model(input_shape,num_classes):
    model = Sequential()
    model.add(Conv2D(32, (3, 3), activation='relu', kernel_initializer='he_uniform', input_shape=input_shape))
    model.add(MaxPool2D((2, 2)))
    model.add(Flatten())
    model.add(Dense(100, activation='relu', kernel_initializer='he_uniform'))
    model.add(Dense(num_classes, activation='softmax'))
  
    opt = SGD(learning_rate=0.01, momentum=0.9)
    model.compile(optimizer=opt, loss='categorical_crossentropy', metrics=['accuracy'])
    
    return model

def setWeightSingleList(weights):
    weights_flat = [w.flatten() for w in weights]    
    weights = np.concatenate(weights_flat).tolist()
    return weights

def reshapeWeight(server_weight, client_weight):
    reshape_weight = []   
    for layer_weights in client_weight:
        n_weights = np.prod(layer_weights.shape)
        reshape_weight.append(np.array(server_weight[:n_weights]).reshape(layer_weights.shape))
        server_weight = server_weight[n_weights:]
    return reshape_weight

def createRandomClientList(clients_dictionary, n_round_clients):
    keys = list(clients_dictionary.keys())
    return random.sample(keys, n_round_clients)


import grpc
import fed_grpc_pb2
import fed_grpc_pb2_grpc
import threading
from concurrent import futures
import queue
import models as aux
import time
import sys
import json  # สำหรับอ่านไฟล์ config

class FedServer(fed_grpc_pb2_grpc.FederatedServiceServicer):
    def __init__(self):
        self.clients = {}
        self.round = 0
        self.avalable_for_register = True

    def sendChunk(self, request_iterator, context):
        for chunk in request_iterator:
            print(f"Received chunk of size: {len(chunk.data)} bytes")
        return fed_grpc_pb2.Response(message="All chunks received successfully.")

    def __sendRound(self):
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            client.sendRound(fed_grpc_pb2.currentRound(round=self.round))

    def __callClientLearning(self, client_ip, q):
        channel = grpc.insecure_channel(client_ip)
        client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
        weight_list = client.startLearning(fed_grpc_pb2.void()).weight
        sample_size = client.getSampleSize(fed_grpc_pb2.void()).size
        q.put([weight_list, sample_size])

    def __callModelValidation(self, aggregated_weights):
        acc_list = []
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            acc_list.append(client.modelValidation(fed_grpc_pb2.weightList(weight=aggregated_weights)).acc)
        return acc_list

    def __FedAvg(self, n_clients, weights_clients_list, sample_size_list):
        if not weights_clients_list or len(weights_clients_list) == 0:
            print("No weights received from clients.")
            return []
        aggregated_weights = []
        for j in range(len(weights_clients_list[0])):
            element = 0.0
            sample_sum = 0.0
            for i in range(n_clients):
                sample_sum += sample_size_list[i]
                element += weights_clients_list[i][j] * sample_size_list[i]
            aggregated_weights.append(element / sample_sum)
        return aggregated_weights

    def killClients(self):
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            client.killClient(fed_grpc_pb2.void())

    def clientRegister(self, request, context):
        ip = request.ip
        port = request.port
        cid = int(request.cid)
        while self.avalable_for_register == False:
            continue
        if cid in self.clients:
            print(f"Could not register Client with ID {cid} - Duplicated Id")
            return fed_grpc_pb2.registerOut(connectedClient=False, round=self.round)
        self.clients[cid] = ip + ":" + port
        print(f"Client {cid} registered!")
        return fed_grpc_pb2.registerOut(connectedClient=True, round=self.round)

    def startServer(self, n_round_clients, min_clients, max_rounds, acc_target):
        while self.round < max_rounds:
            if len(self.clients) < min_clients:
                print("Waiting for the minimum number of clients to connect...")
                while len(self.clients) < min_clients:
                    continue
                print("The minimum number of clients has been reached.")

            self.avalable_for_register = True
            time.sleep(0.5)

            self.round += 1
            self.avalable_for_register = False
            self.__sendRound()
            cid_targets = aux.createRandomClientList(self.clients, n_round_clients)

            thread_list = []
            q = queue.Queue()
            for i in range(n_round_clients):
                thread = threading.Thread(target=self.__callClientLearning, args=(self.clients[cid_targets[i]], q))
                thread_list.append(thread)
                thread.start()
            for thread in thread_list:
                thread.join()

            weights_clients_list = []
            sample_size_list = []
            while not q.empty():
                thread_results = q.get()
                weights_clients_list.append(thread_results[0])
                sample_size_list.append(thread_results[1])

            aggregated_weights = self.__FedAvg(n_round_clients, weights_clients_list, sample_size_list)
            acc_list = self.__callModelValidation(aggregated_weights)
            if not acc_list:
                print("No accuracies to calculate the global accuracy.")
                return

            acc_global = sum(acc_list) / len(acc_list)
            print(f"Round: {self.round} / Accuracy Mean: {acc_global}")
            if acc_global >= acc_target:
                print("Accuracy Target has been achieved! Ending process")
                break

if __name__ == "__main__":
    # Load config from JSON file
    with open('config.json', 'r') as config_file:
        config = json.load(config_file)

    n_round_clients = config["n_round_clients"]
    min_clients = config["min_clients"]
    max_rounds = config["max_rounds"]
    acc_target = config["acc_target"]

    grpc_config = config["grpc_config"]
    ip_address = grpc_config["ip_address"]
    port = grpc_config["port"]
    max_receive_message_length = grpc_config["max_receive_message_length"]

    fed_server = FedServer()

    grpc_server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
        options=[
            ('grpc.max_receive_message_length', max_receive_message_length)
        ]
    )
    
    grpc_address = f'{ip_address}:{port}'
    fed_grpc_pb2_grpc.add_FederatedServiceServicer_to_server(fed_server, grpc_server)
    grpc_server.add_insecure_port(grpc_address)
    grpc_server.start()

    print(f"GRPC server is running on {grpc_address}")

    fed_server.startServer(n_round_clients, min_clients, max_rounds, acc_target)
    fed_server.killClients()


import grpc
import fed_grpc_pb2
import fed_grpc_pb2_grpc
import threading
from concurrent import futures
import queue
import aux as aux
import time
import sys

class FedServer(fed_grpc_pb2_grpc.FederatedServiceServicer):
    def __init__(self):
        self.clients = {}
        self.round = 0
        self.avalable_for_register = True

    def sendChunk(self, request_iterator, context):
        for chunk in request_iterator:
            print(f"Received chunk of size: {len(chunk.data)} bytes")
        return fed_grpc_pb2.Response(message="All chunks received successfully.")

    def __sendRound(self):
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            client.sendRound(fed_grpc_pb2.currentRound(round=(self.round)))

    def __callClientLearning(self, client_ip, q):
        channel = grpc.insecure_channel(client_ip)
        client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
        weight_list = client.startLearning(fed_grpc_pb2.void()).weight
        sample_size = client.getSampleSize(fed_grpc_pb2.void()).size
        q.put([weight_list, sample_size])

    def __callModelValidation(self, aggregated_weights):
        acc_list = []
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            acc_list.append(client.modelValidation(fed_grpc_pb2.weightList(weight=(aggregated_weights))).acc)
        return acc_list

    def __FedAvg(self, n_clients, weights_clients_list, sample_size_list):
        if not weights_clients_list or len(weights_clients_list) == 0:
            print("No weights received from clients.")
            return []
        aggregated_weights = []
        for j in range(len(weights_clients_list[0])):
            element = 0.0
            sample_sum = 0.0
            for i in range(n_clients):
                sample_sum += sample_size_list[i]
                element += weights_clients_list[i][j] * sample_size_list[i]
            aggregated_weights.append(element / sample_sum)
        return aggregated_weights

    def killClients(self):
        for cid in self.clients:
            channel = grpc.insecure_channel(self.clients[cid])
            client = fed_grpc_pb2_grpc.FederatedServiceStub(channel)
            client.killClient(fed_grpc_pb2.void())

    def clientRegister(self, request, context):
        ip = request.ip
        port = request.port
        cid = int(request.cid)
        while self.avalable_for_register == False:
            continue
        if cid in self.clients:
            print(f"Could not register Client with ID {cid} - Duplicated Id")
            return fed_grpc_pb2.registerOut(connectedClient=(False), round=(self.round))
        self.clients[cid] = ip + ":" + port
        print(f"Client {cid} registered!")
        return fed_grpc_pb2.registerOut(connectedClient=(True), round=(self.round))

    def startServer(self, n_round_clients, min_clients, max_rounds, acc_target):
        while self.round < max_rounds:
            if len(self.clients) < min_clients:
                print("Waiting for the minimum number of clients to connect...")
                while len(self.clients) < min_clients:
                    continue
                print("The minimum number of clients has been reached.")
            
            self.avalable_for_register = True
            time.sleep(0.5)

            self.round += 1
            self.avalable_for_register = False

            # ตรวจสอบว่าจำนวน client ไม่มากกว่าจำนวน clients ที่มีอยู่
            if n_round_clients > len(self.clients):
                print(f"Reducing n_round_clients to {len(self.clients)} due to limited available clients.")
                n_round_clients = len(self.clients)

            self.__sendRound()
            cid_targets = aux.createRandomClientList(self.clients, n_round_clients)


            cid_targets = aux.createRandomClientList(self.clients, n_round_clients)
            thread_list = []
            q = queue.Queue()
            for i in range(n_round_clients):
                thread = threading.Thread(target=self.__callClientLearning, args=(self.clients[cid_targets[i]], q))
                thread_list.append(thread)
                thread.start()
            for thread in thread_list:
                thread.join()
            weights_clients_list = []
            sample_size_list = []
            while not q.empty():
                thread_results = q.get()
                weights_clients_list.append(thread_results[0])
                sample_size_list.append(thread_results[1])
            aggregated_weights = self.__FedAvg(n_round_clients, weights_clients_list, sample_size_list)
            acc_list = self.__callModelValidation(aggregated_weights)
            if not acc_list:
                print("No accuracies to calculate the global accuracy.")
                return
            acc_global = sum(acc_list) / len(acc_list)
            print(f"Round: {self.round} / Accuracy Mean: {acc_global}")
            if acc_global >= acc_target:
                print("Accuracy Target has been achieved! Ending process")
                # break

if __name__ == "__main__":
    try:
        n_round_clients = int(sys.argv[1])
        min_clients = int(sys.argv[2])
        max_rounds = int(sys.argv[3])
        acc_target = float(sys.argv[4])
    except IndexError:
        print("Missing argument! You need to pass: (clientsRound, minClients, maxRounds, accuracyTarget)")
        exit()
    fed_server = FedServer()
    grpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=10), options=[('grpc.max_receive_message_length', 1024 * 1024 * 1000)])  # 100 MB
    fed_grpc_pb2_grpc.add_FederatedServiceServicer_to_server(fed_server, grpc_server)
    grpc_server.add_insecure_port('[::]:8081')
    grpc_server.start()
    fed_server.startServer(n_round_clients, min_clients, max_rounds, acc_target)
    fed_server.killClients()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fed_grpc.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'fed_grpc.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x66\x65\x64_grpc.proto\x12\x04main\"F\n\x0cregisterArgs\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\t\x12\x0b\n\x03\x63id\x18\x03 \x01(\t\x12\x0f\n\x07pub_key\x18\x04 \x01(\t\"5\n\x0bregisterOut\x12\x17\n\x0f\x63onnectedClient\x18\x01 \x01(\x08\x12\r\n\x05round\x18\x02 \x01(\x05\"\x1a\n\nsampleSize\x12\x0c\n\x04size\x18\x01 \x01(\x05\"\x17\n\x08\x61\x63\x63uracy\x12\x0b\n\x03\x61\x63\x63\x18\x01 \x01(\x02\"\x1c\n\nweightList\x12\x0e\n\x06weight\x18\x01 \x03(\x02\"\x1d\n\x0c\x63urrentRound\x12\r\n\x05round\x18\x01 \x01(\x05\"\x06\n\x04void\"\x19\n\tChunkData\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\'\n\rEncryptedData\x12\x16\n\x0e\x65ncrypted_data\x18\x01 \x01(\x0c\"\x1b\n\x08Response\x12\x0f\n\x07message\x18\x01 \x01(\t\":\n\x10registerResponse\x12\x17\n\x0f\x63onnectedClient\x18\x01 \x01(\x08\x12\r\n\x05round\x18\x02 \x01(\x05\"<\n\x0e\x45ncryptionArgs\x12*\n\rencryptedData\x18\x01 \x01(\x0b\x32\x13.main.EncryptedData2\xda\x03\n\x10\x46\x65\x64\x65ratedService\x12.\n\tsendChunk\x12\x0f.main.ChunkData\x1a\x0e.main.Response(\x01\x12\x37\n\x0e\x63lientRegister\x12\x12.main.registerArgs\x1a\x11.main.registerOut\x12-\n\rstartLearning\x12\n.main.void\x1a\x10.main.weightList\x12-\n\rgetSampleSize\x12\n.main.void\x1a\x10.main.sampleSize\x12\x33\n\x0fmodelValidation\x12\x10.main.weightList\x1a\x0e.main.accuracy\x12$\n\nkillClient\x12\n.main.void\x1a\n.main.void\x12+\n\tsendRound\x12\x12.main.currentRound\x1a\n.main.void\x12\x39\n\x11sendEncryptedData\x12\x14.main.EncryptionArgs\x1a\x0e.main.Response\x12<\n\x14receiveEncryptedData\x12\x14.main.EncryptionArgs\x1a\x0e.main.Responseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fed_grpc_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_REGISTERARGS']._serialized_start=24
  _globals['_REGISTERARGS']._serialized_end=94
  _globals['_REGISTEROUT']._serialized_start=96
  _globals['_REGISTEROUT']._serialized_end=149
  _globals['_SAMPLESIZE']._serialized_start=151
  _globals['_SAMPLESIZE']._serialized_end=177
  _globals['_ACCURACY']._serialized_start=179
  _globals['_ACCURACY']._serialized_end=202
  _globals['_WEIGHTLIST']._serialized_start=204
  _globals['_WEIGHTLIST']._serialized_end=232
  _globals['_CURRENTROUND']._serialized_start=234
  _globals['_CURRENTROUND']._serialized_end=263
  _globals['_VOID']._serialized_start=265
  _globals['_VOID']._serialized_end=271
  _globals['_CHUNKDATA']._serialized_start=273
  _globals['_CHUNKDATA']._serialized_end=298
  _globals['_ENCRYPTEDDATA']._serialized_start=300
  _globals['_ENCRYPTEDDATA']._serialized_end=339
  _globals['_RESPONSE']._serialized_start=341
  _globals['_RESPONSE']._serialized_end=368
  _globals['_REGISTERRESPONSE']._serialized_start=370
  _globals['_REGISTERRESPONSE']._serialized_end=428
  _globals['_ENCRYPTIONARGS']._serialized_start=430
  _globals['_ENCRYPTIONARGS']._serialized_end=490
  _globals['_FEDERATEDSERVICE']._serialized_start=493
  _globals['_FEDERATEDSERVICE']._serialized_end=967
# @@protoc_insertion_point(module_scope)



import models as aux
import sys
import grpc
import fed_grpc_pb2_grpc
import fed_grpc_pb2
from sklearn.model_selection import train_test_split
from keras.utils import to_categorical
import os
import signal
from concurrent import futures


class FedClient(fed_grpc_pb2_grpc.FederatedServiceServicer):
    def __init__(self, cid, x_train, x_test, y_train, y_test, model, server_adress, client_ip):
        self.cid = cid
        self.x_train = x_train
        self.x_test = x_test
        self.y_train = y_train
        self.y_test = y_test
        self.model = model
        self.server_adress = server_adress
        self.client_ip = client_ip
    
    def __setClientChannel(self,client_channel):
        self.client_channel = client_channel

    
    def __waitingForServer(self,port):
        client_channel = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
        self.__setClientChannel(client_channel)
        fed_grpc_pb2_grpc.add_FederatedServiceServicer_to_server(self, client_channel)

        client_ip = self.client_ip + ':' + port

        client_channel.add_insecure_port(client_ip)
        client_channel.start()
        client_channel.wait_for_termination()

    
    def sendRound(self, request, context):
        ac_round = request.round
        print()
        print(f"Starting {ac_round} round")
        return fed_grpc_pb2.void()

    
    def startLearning(self, request, context):
        self.model.fit(self.x_train, self.y_train, epochs=1, verbose=2)

        weights_list = aux.setWeightSingleList(self.model.get_weights())

        return fed_grpc_pb2.weightList(weight = (weights_list))
    
 
    def getSampleSize(self, request, context):
        return fed_grpc_pb2.sampleSize(size = (len(self.x_train)))

   
    def modelValidation(self, request, context):
        server_weight = request.weight
        self.model.set_weights(aux.reshapeWeight(server_weight, self.model.get_weights()))
        accuracy = self.model.evaluate(self.x_test, self.y_test, verbose=0)[1]

        print(f"Local accuracy with global weights: {accuracy}")

        return fed_grpc_pb2.accuracy(acc = (accuracy))
    
 
    def killClient(self, request, context):
        print()
        print(f"Call for closing channel - Killing Client {self.cid}")
        self.client_channel.stop(0)

        return fed_grpc_pb2.void()
    
    def runClient(self):

        server_channel = grpc.insecure_channel(self.server_adress)
        client = fed_grpc_pb2_grpc.FederatedServiceStub(server_channel)


        port = self.server_adress.split(':')[1]
        port = str(30000 + int(self.cid))

        register_out = client.clientRegister(fed_grpc_pb2.registerArgs(ip=self.client_ip, port=port, cid=self.cid))

        if register_out.connectedClient:
            print(f"Client Connected at round {register_out.round}, waiting for server commands...")
            self.__waitingForServer(port)
        else:
            print("This client couldn't connect with the server")


if __name__ == '__main__':
    cid = -1
    input_shape = (28, 28, 1)
    num_classes = 10
    server_adress = 'localhost:8080'
    client_ip = '[::]'

    try:
        cid = sys.argv[1]
    except IndexError:
        print("Missing argument! You need to pass: ClientId")
        exit()


    x_train, y_train = aux.load_mnist_byCid(cid)
    x_train, x_test, y_train, y_test = train_test_split(x_train, y_train, test_size=0.2, random_state=42)

    y_train = to_categorical(y_train, num_classes)
    y_test = to_categorical(y_test, num_classes)

    model = aux.define_model(input_shape,num_classes)

    fed_client = FedClient(cid, x_train, x_test, y_train, y_test, model, server_adress, client_ip)
    fed_client.runClient()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import fed_grpc_pb2 as fed__grpc__pb2

GRPC_GENERATED_VERSION = '1.66.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in fed_grpc_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class FederatedServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.sendChunk = channel.stream_unary(
                '/main.FederatedService/sendChunk',
                request_serializer=fed__grpc__pb2.ChunkData.SerializeToString,
                response_deserializer=fed__grpc__pb2.Response.FromString,
                _registered_method=True)
        self.clientRegister = channel.unary_unary(
                '/main.FederatedService/clientRegister',
                request_serializer=fed__grpc__pb2.registerArgs.SerializeToString,
                response_deserializer=fed__grpc__pb2.registerOut.FromString,
                _registered_method=True)
        self.startLearning = channel.unary_unary(
                '/main.FederatedService/startLearning',
                request_serializer=fed__grpc__pb2.void.SerializeToString,
                response_deserializer=fed__grpc__pb2.weightList.FromString,
                _registered_method=True)
        self.getSampleSize = channel.unary_unary(
                '/main.FederatedService/getSampleSize',
                request_serializer=fed__grpc__pb2.void.SerializeToString,
                response_deserializer=fed__grpc__pb2.sampleSize.FromString,
                _registered_method=True)
        self.modelValidation = channel.unary_unary(
                '/main.FederatedService/modelValidation',
                request_serializer=fed__grpc__pb2.weightList.SerializeToString,
                response_deserializer=fed__grpc__pb2.accuracy.FromString,
                _registered_method=True)
        self.killClient = channel.unary_unary(
                '/main.FederatedService/killClient',
                request_serializer=fed__grpc__pb2.void.SerializeToString,
                response_deserializer=fed__grpc__pb2.void.FromString,
                _registered_method=True)
        self.sendRound = channel.unary_unary(
                '/main.FederatedService/sendRound',
                request_serializer=fed__grpc__pb2.currentRound.SerializeToString,
                response_deserializer=fed__grpc__pb2.void.FromString,
                _registered_method=True)
        self.sendEncryptedData = channel.unary_unary(
                '/main.FederatedService/sendEncryptedData',
                request_serializer=fed__grpc__pb2.EncryptionArgs.SerializeToString,
                response_deserializer=fed__grpc__pb2.Response.FromString,
                _registered_method=True)
        self.receiveEncryptedData = channel.unary_unary(
                '/main.FederatedService/receiveEncryptedData',
                request_serializer=fed__grpc__pb2.EncryptionArgs.SerializeToString,
                response_deserializer=fed__grpc__pb2.Response.FromString,
                _registered_method=True)


class FederatedServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def sendChunk(self, request_iterator, context):
        """Streaming RPC สำหรับการส่งข้อมูลขนาดใหญ่จากไคลเอนต์ไปยังเซิร์ฟเวอร์
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def clientRegister(self, request, context):
        """RPC สำหรับการลงทะเบียน client
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def startLearning(self, request, context):
        """RPC เพื่อเริ่มการเรียนรู้ และส่งกลับรายการ weights
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getSampleSize(self, request, context):
        """RPC เพื่อรับขนาดตัวอย่างจากเซิร์ฟเวอร์
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def modelValidation(self, request, context):
        """RPC สำหรับการตรวจสอบความถูกต้องของโมเดล (validation)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def killClient(self, request, context):
        """RPC เพื่อหยุด client
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendRound(self, request, context):
        """RPC เพื่อส่งรอบปัจจุบัน
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendEncryptedData(self, request, context):
        """RPC สำหรับส่งข้อมูลที่เข้ารหัสจาก client ไปยัง server
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def receiveEncryptedData(self, request, context):
        """RPC สำหรับรับข้อมูลที่เข้ารหัสจาก server
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FederatedServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'sendChunk': grpc.stream_unary_rpc_method_handler(
                    servicer.sendChunk,
                    request_deserializer=fed__grpc__pb2.ChunkData.FromString,
                    response_serializer=fed__grpc__pb2.Response.SerializeToString,
            ),
            'clientRegister': grpc.unary_unary_rpc_method_handler(
                    servicer.clientRegister,
                    request_deserializer=fed__grpc__pb2.registerArgs.FromString,
                    response_serializer=fed__grpc__pb2.registerOut.SerializeToString,
            ),
            'startLearning': grpc.unary_unary_rpc_method_handler(
                    servicer.startLearning,
                    request_deserializer=fed__grpc__pb2.void.FromString,
                    response_serializer=fed__grpc__pb2.weightList.SerializeToString,
            ),
            'getSampleSize': grpc.unary_unary_rpc_method_handler(
                    servicer.getSampleSize,
                    request_deserializer=fed__grpc__pb2.void.FromString,
                    response_serializer=fed__grpc__pb2.sampleSize.SerializeToString,
            ),
            'modelValidation': grpc.unary_unary_rpc_method_handler(
                    servicer.modelValidation,
                    request_deserializer=fed__grpc__pb2.weightList.FromString,
                    response_serializer=fed__grpc__pb2.accuracy.SerializeToString,
            ),
            'killClient': grpc.unary_unary_rpc_method_handler(
                    servicer.killClient,
                    request_deserializer=fed__grpc__pb2.void.FromString,
                    response_serializer=fed__grpc__pb2.void.SerializeToString,
            ),
            'sendRound': grpc.unary_unary_rpc_method_handler(
                    servicer.sendRound,
                    request_deserializer=fed__grpc__pb2.currentRound.FromString,
                    response_serializer=fed__grpc__pb2.void.SerializeToString,
            ),
            'sendEncryptedData': grpc.unary_unary_rpc_method_handler(
                    servicer.sendEncryptedData,
                    request_deserializer=fed__grpc__pb2.EncryptionArgs.FromString,
                    response_serializer=fed__grpc__pb2.Response.SerializeToString,
            ),
            'receiveEncryptedData': grpc.unary_unary_rpc_method_handler(
                    servicer.receiveEncryptedData,
                    request_deserializer=fed__grpc__pb2.EncryptionArgs.FromString,
                    response_serializer=fed__grpc__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'main.FederatedService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('main.FederatedService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class FederatedService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def sendChunk(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/main.FederatedService/sendChunk',
            fed__grpc__pb2.ChunkData.SerializeToString,
            fed__grpc__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def clientRegister(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/clientRegister',
            fed__grpc__pb2.registerArgs.SerializeToString,
            fed__grpc__pb2.registerOut.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def startLearning(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/startLearning',
            fed__grpc__pb2.void.SerializeToString,
            fed__grpc__pb2.weightList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getSampleSize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/getSampleSize',
            fed__grpc__pb2.void.SerializeToString,
            fed__grpc__pb2.sampleSize.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def modelValidation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/modelValidation',
            fed__grpc__pb2.weightList.SerializeToString,
            fed__grpc__pb2.accuracy.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def killClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/killClient',
            fed__grpc__pb2.void.SerializeToString,
            fed__grpc__pb2.void.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def sendRound(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/sendRound',
            fed__grpc__pb2.currentRound.SerializeToString,
            fed__grpc__pb2.void.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def sendEncryptedData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/sendEncryptedData',
            fed__grpc__pb2.EncryptionArgs.SerializeToString,
            fed__grpc__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def receiveEncryptedData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/main.FederatedService/receiveEncryptedData',
            fed__grpc__pb2.EncryptionArgs.SerializeToString,
            fed__grpc__pb2.Response.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import urllib.request
import os
import gzip
from PIL import Image
import numpy as np
import random
import shutil
import sys

def deleteAllFolder(path):
    for file_name in os.listdir(path):
        file_path = os.path.join(path, file_name)
        if os.path.isfile(file_path):
            os.remove(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)

def downloadSaveData():
   
    train_data_url = 'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz'
    train_labels_url = 'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz'

   
    train_data_file = 'train-images-idx3-ubyte.gz'
    train_labels_file = 'train-labels-idx1-ubyte.gz'

   
    data_dir = 'data'

   
    if os.path.exists(data_dir):
        deleteAllFolder(data_dir)
    else:
        
        os.makedirs(data_dir)

   
    urllib.request.urlretrieve(train_data_url, os.path.join(data_dir, train_data_file))
    urllib.request.urlretrieve(train_labels_url, os.path.join(data_dir, train_labels_file))

  
    with gzip.open(os.path.join(data_dir, train_data_file), 'rb') as f:
        train_images = np.frombuffer(f.read(), dtype=np.uint8, offset=16).reshape((-1, 28, 28))
    with gzip.open(os.path.join(data_dir, train_labels_file), 'rb') as f:
        train_labels = np.frombuffer(f.read(), dtype=np.uint8, offset=8)

   
    out_dir = 'data/images'

    for i, (img, label) in enumerate(zip(train_images, train_labels)):
        img = Image.fromarray(img, mode='L')
        label_dir = os.path.join(out_dir, str(label))
        os.makedirs(label_dir, exist_ok=True)
        os.makedirs(label_dir, exist_ok=True)
        img.save(os.path.join(label_dir, f'{i}.jpg'))

def split_data(input_dir, output_dir, n_clients):
   
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    for label in range(10):
       
        file_names = os.listdir(input_dir+'/'+str(label))
        
       
        random.shuffle(file_names)
        
       
        num_files_per_partition = len(file_names) // n_clients
        
        for i in range(n_clients):
          
            start_index = i * num_files_per_partition
            end_index = start_index + num_files_per_partition
            
          
            if i == n_clients - 1:
                end_index = len(file_names)
            
            
            if label == 0:
                client_dir = os.path.join(output_dir, f'client_{i}')
                os.makedirs(client_dir, exist_ok=True)

            client_dir_label = os.path.join(output_dir, f'client_{i}/{label}')
            os.makedirs(client_dir_label, exist_ok=True)
            
            
            for file_name in file_names[start_index:end_index]:
                src_path = os.path.join(input_dir+'/'+str(label), file_name)
                dst_path = os.path.join(client_dir_label, file_name)
                shutil.copy(src_path, dst_path)

    deleteAllFolder(input_dir)

if __name__ == '__main__':
    try:
        n_clients = int(sys.argv[1])
    except IOError:
        print("Missing argument! Number of clients...")
        exit()

    downloadSaveData()
    split_data("data/images", "data", n_clients)


import sys
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    readme = fh.read()

with open("CHANGELOG.md", "r", encoding="utf-8") as fh:
    changelog = fh.read()

long_description = readme + "\n\n" + changelog

install_requires = [
    'aiosignal==1.3.1',
    'anyio==4.3.0',
    'argon2-cffi==23.1.0',
    'argon2-cffi-bindings==21.2.0',
    'arrow==1.3.0',
    # ... (other packages) ...
    'grpcio==1.66.1',
    'imutils',
    'keras',
    'h11==0.14.0',
    'httpcore==1.0.5',
    'httpx==0.27.0',
    'idna==3.6',
    'ipykernel',
    'ipython',
    'isoduration==20.11.0',
    'iterators==0.0.2',
    'jedi==0.19.1',
    'Jinja2==3.1.3',
    'json5==0.9.25',
    'jsonpointer==2.4',
    'jsonschema==4.21.1',
    'jsonschema-specifications==2023.12.1',
    'jupyter-lsp',
    'jupyter_client',
    'jupyter_core',
    'jupyter_server',
    'jupyter_server_terminals',
    'jupyterlab',
    'jupyterlab_pygments',
    'jupyterlab_server',
    'kiwisolver==1.4.5',
    'markdown-it-py',
    'MarkupSafe==2.1.4',
    'matplotlib',
    'matplotlib-inline==0.1.7',
    'mdurl==0.1.2',
    'mistune==3.0.2',
    'mpmath==1.3.0',
    'msgpack==1.0.7',
    'nbclient==0.10.0',
    'nbconvert==7.16.4',
    'nbformat==5.10.4',
    'nest-asyncio==1.6.0',
    'networkx',
    'notebook_shim==0.2.4',
    'numpy',
    'overrides==7.7.0',
    'packaging==23.2',
    'pandas',
    'pandocfilters==1.5.1',
    'parso==0.8.4',
    'pexpect==4.9.0',
    'pillow==10.2.0',
    'protobuf',
    'platformdirs==4.2.2',
    'prometheus_client==0.20.0',
    'prompt-toolkit==3.0.43',
    'psutil==5.9.8',
    'ptyprocess==0.7.0',
    'pure-eval==0.2.2',
    'py-markdown-table==0.4.0',
    'pyarrow',
    'pybind11-global==2.11.1',
    'pycparser==2.21',
    'pycryptodome==3.20.0',
    'Pygments==2.18.0',
    'pyparsing==3.1.1',
    'python-dateutil==2.8.2',
    'python-json-logger==2.0.7',
    'pytz==2023.3.post1',
    'PyYAML',
    'referencing==0.32.1',
    'requests==2.31.0',
    'rfc3339-validator==0.1.4',
    'rfc3986-validator==0.1.1',
    'rich==13.7.1',
    'rpds-py==0.17.1',
    'Send2Trash==1.8.3',
    'shellingham==1.5.4',
    'six==1.16.0',
    'sniffio==1.3.1',
    'soupsieve==2.5',
    'stack-data==0.6.3',
    'sympy==1.12',
    'terminado==0.18.1',
    'tinycss2==1.3.0',
    'tomli==2.0.1',
    'torch',
    'torchvision',
    'tornado',
    'tqdm',
    'traitlets==5.14.3',
    'typer==0.9.4',
    'types-python-dateutil==2.9.0.20240316',
    'typing_extensions',
    'tzdata==2023.4',
    'uri-template==1.3.0',
    'urllib3==2.1.0',
    'wcwidth==0.2.13',
    'webcolors==1.13',
    'webencodings==0.5.1',
    'websocket-client==1.8.0',
]

# Add conditional TensorFlow installation based on platform
if sys.platform == "darwin":  # macOS
    install_requires.append('tensorflow-macos==2.13.0')
    install_requires.append('tensorflow-metal')
else:  # Other platforms, such as Linux (e.g., Colab)
    install_requires.append('tensorflow==2.13.0')

setup(
    name="cifer",
    version="0.1.19",
    author="cifer.ai",
    author_email="parit@cifer.ai",
    description="Federated Learning and Fully Homomorphic Encryption",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.cifer.ai",
    license="Apache License 2.0",
    packages=find_packages(),
    package_dir={'client': 'Client'},
    install_requires=install_requires,  # Use the conditional install_requires
    tests_require=[
        'coverage', 'wheel', 'pytest', 'requests_mock'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",  
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",  
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ]
)



from keras.models import Sequential
from keras.layers import Conv2D, MaxPooling2D, Flatten, Dense
from keras.optimizers import Adam
from keras.datasets import mnist
import numpy as np
import random

def define_model(input_shape, num_classes):
    model = Sequential()
    model.add(Conv2D(32, kernel_size=(3, 3), activation='relu', input_shape=input_shape))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(64, kernel_size=(3, 3), activation='relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(128, activation='relu'))
    model.add(Dense(num_classes, activation='softmax'))
    model.compile(optimizer=Adam(), loss='categorical_crossentropy', metrics=['accuracy'])
    return model

def load_mnist_byCid(cid):
    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train = x_train.astype('float32') / 255.0
    x_test = x_test.astype('float32') / 255.0
    x_train = np.expand_dims(x_train, -1)
    x_test = np.expand_dims(x_test, -1)
    client_data_size = len(x_train) // 5
    start_idx = cid * client_data_size
    end_idx = (cid + 1) * client_data_size
    return x_train[start_idx:end_idx], y_train[start_idx:end_idx]

def setWeightSingleList(weight_list):
    flattened_weights = []
    for layer_weights in weight_list:
        flattened_weights.extend(layer_weights.flatten())
    return flattened_weights

def reshapeWeight(flat_weights, weight_shapes):
    reshaped_weights = []
    current_position = 0
    for layer_weights in weight_shapes:
        shape_size = np.prod(layer_weights.shape)
        reshaped_weights.append(np.array(flat_weights[current_position:current_position + shape_size]).reshape(layer_weights.shape))
        current_position += shape_size
    return reshaped_weights

def createRandomClientList(clients, n_round_clients):
    return random.sample(list(clients.keys()), n_round_clients)



from lightphe import LightPHE

class HomomorphicEncryption:
    def __init__(self, algorithm_name="Paillier"):
        # สร้างคู่กุญแจส่วนตัว-สาธารณะ
        self.cs = LightPHE(algorithm_name=algorithm_name)
        self.cs.export_keys(target_file="public.txt", public=True)
        self.cs.export_keys(target_file="private.txt", public=False)

    def encrypt(self, message):
        """ฟังก์ชันการเข้ารหัส"""
        return self.cs.encrypt(message)

    def decrypt(self, ciphertext):
        """ฟังก์ชันการถอดรหัส"""
        return self.cs.decrypt(ciphertext)

# ตัวอย่างการใช้งาน
if __name__ == "__main__":
    # สร้างอ็อบเจ็กต์ Homomorphic
    homomorphic_encryption = HomomorphicEncryption()

    # กำหนดข้อความธรรมชาติ
    m1 = 17
    m2 = 23

    # เข้ารหัสข้อความ
    c1 = homomorphic_encryption.encrypt(m1)
    c2 = homomorphic_encryption.encrypt(m2)

    print("Ciphertext 1:", c1.value)  # แสดง ciphertext
    print("Ciphertext 2:", c2.value)  # แสดง ciphertext

    # ถอดรหัสข้อความ
    dec_m1 = homomorphic_encryption.decrypt(c1)
    dec_m2 = homomorphic_encryption.decrypt(c2)

    print("Decrypted message 1:", dec_m1)  # ควรแสดง 17
    print("Decrypted message 2:", dec_m2)  # ควรแสดง 23
